#ifndef FSLBASE_MATH_MATHHELPER_HPP
#define FSLBASE_MATH_MATHHELPER_HPP
/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <algorithm>
#include <FslBase/Math/Point2.hpp>
#include <FslBase/Math/RectangleSizeRestrictionFlag.hpp>

namespace Fsl
{
  namespace MathHelper
  {
    const float PI = 3.1415926535897932384626433832795028841971693f;
    //! PI / 180.0f
    const float TO_RADS = 0.017453292519943295769236907684886f;
    //! 1.0f / (PI / 180.0f)
    const float TO_DEGREES = 57.295779513082320876798154814105f;


    //! @brief Restricts a value to be within a specified range.
    inline float Clamp(const float value, const float min, const float max)
    {
      return std::min(std::max(value, min), max);
    }

    //! @brief Linearly interpolates between two values.
    inline float Lerp(const float value1, const float value2, const float amount)
    {
      return value1 + ((value2 - value1) * amount);
    }

    //! @brief Converts radians to degrees.
    //! @param radians The angle in radians.
    //! @return The angle in degrees.
    inline float ToDegrees(const float radians)
    {
      // We use doubles to not loose precision
      return static_cast<float>(radians * 57.295779513082320876798154814105);
    }

    //! @brief Converts degrees to radians.
    //! @param degrees The angle in degrees.
    //! @return The angle in radians.
    inline float ToRadians(const float degrees)
    {
      // We use doubles to not loose precision
      return static_cast<float>(degrees * 0.017453292519943295769236907684886);
    }

    //! @brief Find the nearest power of two value that is greater or equal the input value (>=)
    //! @param value must be >= 0;
    extern int ToPowerOfTwo(const int value);

    //! @brief Find the optimal rectangle size for packing a given amount of uniform sized units under the supplied constraints
    //! @param unitSize the size of the unit (x > 0 && y > 0)
    //! @param unitCount the total number of units > 0
    extern Point2 CalcOptimalSize(const Point2& unitSize, const int32_t unitCount, const RectangleSizeRestrictionFlag::Enum restrictionFlags);
  };
}

#endif

