/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslGraphics/Exceptions.hpp>
#include <FslGraphics/PixelFormatUtil.hpp>
#include <cassert>

namespace Fsl
{
  namespace
  {
    const int32_t MASK_FORMAT_ID = 0x00FFFF;
    const int32_t MASK_BYTES_PER_PIXEL = 0x0F0000;
    const int32_t SHIFT_BYTES_PER_PIXEL = 16;
  }

  int32_t PixelFormatUtil::GetBytesPerPixel(const PixelFormat::Enum pixelFormat)
  {
    return (pixelFormat & MASK_BYTES_PER_PIXEL) >> SHIFT_BYTES_PER_PIXEL;
  }


  int32_t PixelFormatUtil::CalcMinimumStride(const int32_t width, const PixelFormat::Enum pixelFormat)
  {
    return CalcMinimumStride(width, GetBytesPerPixel(pixelFormat));
  }


  int32_t PixelFormatUtil::CalcMinimumStride(const int32_t width, const int32_t bytesPerPixel)
  {
    assert(width >= 0);
    assert(bytesPerPixel >= 0);

    return width * bytesPerPixel;
  }


  int32_t PixelFormatUtil::CalcMinimumStride(const int32_t width, const PixelFormat::Enum pixelFormat, const StrideRequirement::Enum strideRequirement)
  {
    return CalcMinimumStride(width, GetBytesPerPixel(pixelFormat), strideRequirement);
  }


  int32_t PixelFormatUtil::CalcMinimumStride(const int32_t width, const int32_t bytesPerPixel, const StrideRequirement::Enum strideRequirement)
  {
    assert(width >= 0);
    assert(bytesPerPixel >= 0);

    int32_t minStride = width * bytesPerPixel;
    switch (strideRequirement)
    {
    case StrideRequirement::Minimum:
    case StrideRequirement::Any:
      return minStride;
    case StrideRequirement::MinimumAlign16Bit:
    case StrideRequirement::Align16Bit:
      return minStride + (minStride & 1);
    case StrideRequirement::MinimumAlign32Bit:
    case StrideRequirement::Align32Bit:
      return minStride + ((minStride & 3) == 0 ? 0 : 4 - (minStride & 3));
    default:
      throw NotSupportedException("Unknown StrideRequirement");
    }
  }


  bool PixelFormatUtil::IsValidStride(const int32_t width, const PixelFormat::Enum pixelFormat, const StrideRequirement::Enum strideRequirement, const int32_t desiredStride)
  {
    return IsValidStride(width, GetBytesPerPixel(pixelFormat), strideRequirement, desiredStride);
  }


  bool PixelFormatUtil::IsValidStride(const int32_t width, const int32_t bytesPerPixel, const StrideRequirement::Enum strideRequirement, const int32_t desiredStride)
  {
    assert(width >= 0);
    assert(bytesPerPixel >= 0);

    const int32_t minStride = width * bytesPerPixel;
    switch (strideRequirement)
    {
    case StrideRequirement::Any:
      return desiredStride >= minStride;
    case StrideRequirement::Minimum:
      return desiredStride == minStride;
    case StrideRequirement::MinimumAlign16Bit:
      return desiredStride == (minStride + (minStride & 1));
    case StrideRequirement::MinimumAlign32Bit:
      return desiredStride == (minStride + ((minStride & 3) == 0 ? 0 : 4 - (minStride & 3)));
    case StrideRequirement::Align16Bit:
      return (desiredStride & 1) == 0;
    case StrideRequirement::Align32Bit:
      return (desiredStride & 3) == 0;
    default:
      throw NotSupportedException("Unknown StrideRequirement");
    }
  }


}
