/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslGraphics/Exceptions.hpp>
#include <FslGraphics/PixelFormatUtil.hpp>
#include <FslGraphics/RawBitmapUtil.hpp>
#include <FslGraphics/RawBitmapEx.hpp>
#include <cassert>

namespace Fsl
{


  int RawBitmapUtil::CalcAlignment(const RawBitmap& rawBitmap, const int bytesPerPixel)
  {
    assert(rawBitmap.Stride() > 0);

    const int minStride = rawBitmap.Width() * bytesPerPixel;
    const int stride = rawBitmap.Stride();
    const int minStride8 = minStride + (((minStride & 7) != 0 ? (8 - (minStride & 7)) : 0));
    const int minStride4 = minStride + (((minStride & 3) != 0 ? (4 - (minStride & 3)) : 0));
    const int minStride2 = minStride + (minStride & 1);

    // Check for 8 byte alignment
    if ((stride & 7) == 0 && stride == minStride8)
      return 8;
    else if ((stride & 3) == 0 && stride == minStride4)
      return 4;
    else if ((stride & 1) == 0 && stride == minStride2)
      return 2;
    else
    {
      if (stride != minStride)
        throw UnsupportedStrideException("Stride does not fit with any supported alignment", stride);
      return 1;
    }
  }



  void RawBitmapUtil::Swizzle(RawBitmapEx& rBitmap, const int srcIdx0, const int srcIdx1, const int srcIdx2, const int srcIdx3)
  {
    if (!rBitmap.IsValid())
      throw std::invalid_argument("Swizzle requires a valid bitmap");

    switch (rBitmap.GetPixelFormat())
    {
    case PixelFormat::RGB888:
    case PixelFormat::BGR888:
      Swizzle24(rBitmap, srcIdx0, srcIdx1, srcIdx2);
      break;
    case PixelFormat::RGBA8888:
    case PixelFormat::BGRA8888:
      Swizzle32(rBitmap, srcIdx0, srcIdx1, srcIdx2, srcIdx3);
      break;
    default:
      throw UnsupportedPixelFormatException("Swizzle only supports RGB888, BGR888, RGBA8888 or BGRA8888 format", rBitmap.GetPixelFormat());
    }
  }


  void RawBitmapUtil::Swizzle24(RawBitmapEx& rBitmap, const int srcIdx0, const int srcIdx1, const int srcIdx2)
  {
    if (!rBitmap.IsValid())
      throw std::invalid_argument("Swizzle requires a valid bitmap");
    if( rBitmap.GetPixelFormat() != PixelFormat::RGB888 && rBitmap.GetPixelFormat() != PixelFormat::BGR888)
      throw UnsupportedPixelFormatException("Swizzle24 only supports RGB888 or BGR888 format", rBitmap.GetPixelFormat());

    assert(srcIdx0 >= 0 && srcIdx0 <= 2);
    assert(srcIdx1 >= 0 && srcIdx1 <= 2);
    assert(srcIdx2 >= 0 && srcIdx2 <= 2);
    assert(PixelFormatUtil::GetBytesPerPixel(rBitmap.GetPixelFormat()) == 3);


    // Generic slow swizzle for 24bpp formats
    uint8_t* pBitmap = static_cast<uint8_t*>(rBitmap.Content());
    const int srcWidth = rBitmap.Width();
    const int srcHeight = rBitmap.Height();
    const int srcStride = rBitmap.Stride();
    int offset = 0;
    for (int y = 0; y < srcHeight; ++y)
    {
      for (int x = 0; x < srcWidth; ++x)
      {
        const uint8_t b0 = pBitmap[offset + (x * 3) + srcIdx0];
        const uint8_t b1 = pBitmap[offset + (x * 3) + srcIdx1];
        const uint8_t b2 = pBitmap[offset + (x * 3) + srcIdx2];
        pBitmap[offset + (x * 3) + 0] = b0;
        pBitmap[offset + (x * 3) + 1] = b1;
        pBitmap[offset + (x * 3) + 2] = b2;
      }
      offset += srcStride;
    }
  }


  void RawBitmapUtil::Swizzle32(RawBitmapEx& rBitmap, const int srcIdx0, const int srcIdx1, const int srcIdx2, const int srcIdx3)
  {
    if (!rBitmap.IsValid())
      throw std::invalid_argument("Swizzle requires a valid bitmap");
    if (rBitmap.GetPixelFormat() != PixelFormat::RGBA8888 && rBitmap.GetPixelFormat() != PixelFormat::BGRA8888)
      throw UnsupportedPixelFormatException("Swizzle32 only supports RGBA8888 or BGRA8888 format", rBitmap.GetPixelFormat());

    assert(srcIdx0 >= 0 && srcIdx0 <= 3);
    assert(srcIdx1 >= 0 && srcIdx1 <= 3);
    assert(srcIdx2 >= 0 && srcIdx2 <= 3);
    assert(srcIdx3 >= 0 && srcIdx3 <= 3);
    assert(PixelFormatUtil::GetBytesPerPixel(rBitmap.GetPixelFormat()) == 4);


    // Generic slow swizzle for 32bpp formats
    uint8_t* pBitmap = static_cast<uint8_t*>(rBitmap.Content());
    const int srcWidth = rBitmap.Width();
    const int srcHeight = rBitmap.Height();
    const int srcStride = rBitmap.Stride();
    int offset = 0;
    for (int y = 0; y < srcHeight; ++y)
    {
      for (int x = 0; x < srcWidth; ++x)
      {
        const uint8_t b0 = pBitmap[offset + (x * 4) + srcIdx0];
        const uint8_t b1 = pBitmap[offset + (x * 4) + srcIdx1];
        const uint8_t b2 = pBitmap[offset + (x * 4) + srcIdx2];
        const uint8_t b3 = pBitmap[offset + (x * 4) + srcIdx3];
        pBitmap[offset + (x * 4) + 0] = b0;
        pBitmap[offset + (x * 4) + 1] = b1;
        pBitmap[offset + (x * 4) + 2] = b2;
        pBitmap[offset + (x * 4) + 3] = b3;
      }
      offset += srcStride;
    }
  }


  void RawBitmapUtil::ClearPadding(RawBitmapEx& rBitmap)
  {
    if (!rBitmap.IsValid())
      throw std::invalid_argument("ClearPadding requires a valid bitmap");

    // Generic padding clear
    const int32_t minimumStride = PixelFormatUtil::CalcMinimumStride(rBitmap.Width(), rBitmap.GetPixelFormat());
    const int32_t stride = rBitmap.Stride();
    const int32_t paddingSize = stride - minimumStride;

    assert(paddingSize >= 0);
    if (paddingSize > 0)
    {
      uint8_t* pDst = static_cast<uint8_t*>(rBitmap.Content());
      const uint8_t*const pDstEnd = pDst + (rBitmap.Height() * stride);

      // Move to the padding area
      pDst += minimumStride;
      const uint8_t* pDstPaddingEnd = pDst + paddingSize;

      while (pDst < pDstEnd)
      {
        // Clear the padding area
        while (pDst < pDstPaddingEnd)
        {
          *pDst = 0;
          ++pDst;
        }
        pDst += minimumStride;
        pDstPaddingEnd += stride;
      }
    }
  }



  void RawBitmapUtil::FlipHorizontal(RawBitmapEx& rBitmap)
  {
    if (!rBitmap.IsValid())
      throw std::invalid_argument("Swizzle requires a valid bitmap");

    // Generic slow flip
    const int32_t minimumStride = PixelFormatUtil::CalcMinimumStride(rBitmap.Width(), rBitmap.GetPixelFormat());

    const int32_t dstStride = rBitmap.Stride();
    uint8_t* pDst = static_cast<uint8_t*>(rBitmap.Content());
    const uint8_t*const pDstEnd = pDst + (dstStride * (rBitmap.Height() / 2));

    uint8_t* pSrc = pDst + (dstStride * (rBitmap.Height() - 1));

    uint8_t tmp;

    const int32_t dstStrideAdjust = dstStride - minimumStride;
    const int32_t srcStride = dstStride + minimumStride;

    uint8_t* pDstTmp = pDst + minimumStride;
    while (pDst < pDstEnd)
    {
      while (pDst < pDstTmp)
      {
        tmp = *pDst;
        *pDst = *pSrc;
        *pSrc = tmp;
        ++pSrc;
        ++pDst;
      }
      pSrc -= srcStride;
      pDst += dstStrideAdjust;
      pDstTmp += dstStride;
    }

  }

}
