/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslGraphicsGLES2/Exceptions.hpp>
#include <FslGraphicsGLES2/GLCheck.hpp>
#include <FslGraphicsGLES2/GLVertexBuffer.hpp>

#include <algorithm>
#include <cassert>

namespace Fsl
{
  namespace GLES2
  {
    namespace
    {

      // Convert from the platform independent vertex declaration to a GL one
      void ConvertToNative(std::vector<GLVertexElement>& rVertexElements, const VertexDeclaration& vertexDeclaration)
      {
        assert(vertexDeclaration.Count() >= 0);
        if (size_t(vertexDeclaration.Count()) != rVertexElements.size())
          rVertexElements.resize(vertexDeclaration.Count());

        const int srcCount = vertexDeclaration.Count();
        const VertexElementEx* pSrcElements = vertexDeclaration.DirectAccess();
        const VertexElementEx*const pSrcElementsEnd = pSrcElements + srcCount;
        GLVertexElement* pDst = rVertexElements.data();

        while (pSrcElements < pSrcElementsEnd)
        {
          pDst->Reset(*pSrcElements);
          ++pSrcElements;
          ++pDst;
        }
      }
    }


    GLVertexBuffer::GLVertexBuffer()
      : GLBuffer()
      , m_vertexElements()
    {
    }


    GLVertexBuffer::GLVertexBuffer(const void*const pVertices, const int elementCount, const VertexDeclaration& vertexDeclaration, const GLenum usage)
      : GLBuffer(GL_ARRAY_BUFFER, pVertices, elementCount, vertexDeclaration.VertexStride(), usage)
      , m_vertexElements()
    {
      ConvertToNative(m_vertexElements, vertexDeclaration);
    }


    void GLVertexBuffer::Reset(const void*const pVertices, const int elementCount, const VertexDeclaration& vertexDeclaration, const GLenum usage)
    {
      GLBuffer::Reset(GL_ARRAY_BUFFER, pVertices, elementCount, vertexDeclaration.VertexStride(), usage);

      ConvertToNative(m_vertexElements, vertexDeclaration);
    }


    void GLVertexBuffer::EnableAttribArrays()
    {
      const int vertexStride = GetElementStride();
      GLVertexElement* pVertexElements = m_vertexElements.data();
      const GLVertexElement*const pVertexElementsEnd = pVertexElements + m_vertexElements.size();

      GLuint index = 0;
      while (pVertexElements < pVertexElementsEnd)
      {
        glVertexAttribPointer(index, pVertexElements->Size, pVertexElements->Type, pVertexElements->Normalized, vertexStride, pVertexElements->Pointer);
        glEnableVertexAttribArray(index);
        ++pVertexElements;
        ++index;
      }
    }


    void GLVertexBuffer::DisableAttribArrays()
    {
      const GLuint count = m_vertexElements.size();
      GLuint index = 0;
      while (index < count)
      {
        glDisableVertexAttribArray(index);
        ++index;
      }
    }


    void GLVertexBuffer::EnableAttribArrays(const GLuint*const pAttributeIndices, const int count)
    {
      const int vertexStride = GetElementStride();
      GLVertexElement* pVertexElements = m_vertexElements.data();
      const GLVertexElement*const pVertexElementsEnd = pVertexElements + m_vertexElements.size();

      const GLuint* pAttributeIndex = pAttributeIndices;
      const GLuint*const pAttributeIndexEnd = pAttributeIndices + count;
      while (pVertexElements < pVertexElementsEnd && pAttributeIndex < pAttributeIndexEnd)
      {
        GLuint index = *pAttributeIndex;
        glVertexAttribPointer(index, pVertexElements->Size, pVertexElements->Type, pVertexElements->Normalized, vertexStride, pVertexElements->Pointer);
        glEnableVertexAttribArray(index);
        ++pVertexElements;
        ++pAttributeIndex;
      }
    }


    void GLVertexBuffer::DisableAttribArrays(const GLuint*const pAttributeIndices, const int count)
    {
      for (int i = 0; i < count; ++i)
      {
        glDisableVertexAttribArray(pAttributeIndices[i]);
      }
    }


    void GLVertexBuffer::EnableAttribArrays(const GLVertexAttribLink*const pLinks, const int count)
    {
      const int vertexStride = GetElementStride();
      for (int i = 0; i < count; ++i)
      {
        const int attribIndex = pLinks[i].AttribIndex;
        const int elementIndex = pLinks[i].VertexElementIndex;
        assert(elementIndex >= 0);
        assert(elementIndex < int(m_vertexElements.size()));
        glVertexAttribPointer(attribIndex, m_vertexElements[elementIndex].Size, m_vertexElements[elementIndex].Type,
          m_vertexElements[elementIndex].Normalized, vertexStride, m_vertexElements[elementIndex].Pointer);
        glEnableVertexAttribArray(attribIndex);
      }
    }


    void GLVertexBuffer::DisableAttribArrays(const GLVertexAttribLink*const pLinks, const int count)
    {
      for (int i = 0; i < count; ++i)
      {
        glDisableVertexAttribArray(pLinks[i].AttribIndex);
      }
    }


    int GLVertexBuffer::VertexAttribCount() const
    {
      return m_vertexElements.size();
    }

    const GLVertexElement* GLVertexBuffer::VertexAttribDirectAccess() const
    {
      return m_vertexElements.data();
    }


    int GLVertexBuffer::GetVertexElementIndex(const VertexElementUsage::Enum usage, const int usageIndex) const
    {
      const int index = VertexElementIndexOf(usage, usageIndex);
      if (index < 0)
        throw UsageErrorException("The supplied usage and usageIndex combo was not found");
      return index;
    }


    int GLVertexBuffer::VertexElementIndexOf(const VertexElementUsage::Enum usage, const int usageIndex) const
    {
      for (size_t i = 0; i < m_vertexElements.size(); ++i)
      {
        if (m_vertexElements[i].Source.Usage == usage && m_vertexElements[i].Source.UsageIndex == usageIndex)
          return i;
      }
      return -1;
    }
  }
}
